@@ -1,4 +1,4 @@
-# $Id: UserAgent.pm,v 2.0 2001/11/19 16:55:54 gisle Exp $
+# $Id: UserAgent.pm,v 2.4 2003/02/04 16:36:55 gisle Exp $
 
 package LWP::UserAgent;
 use strict;
@@ -16,16 +16,30 @@ LWP::UserAgent - A WWW UserAgent class
                              );
 
  $response = $ua->get('http://search.cpan.org/');
+ die "Error while getting ", $response->request->uri,
+   " -- ", $response->status_line, "\nAborting"
+  unless $response->is_success;
 
  # or:
 
+ $response = $ua->get('http://search.cpan.org/',
+                        ':content_file' => '/tmp/sco.html'
+                     );
+ # or:
+
+ $response = $ua->get('http://search.cpan.org/',
+                        ':content_cb'     => \&callback,
+                        ':read_size_hint' => 4096,
+                     );
+ # or:
+
  $request = HTTP::Request->new('GET', 'http://search.cpan.org/');
   # and then one of these:
  $response = $ua->request($request); # or
- $response = $ua->request($request, '/tmp/sss'); # or
+ $response = $ua->request($request, '/tmp/sco.html'); # or
  $response = $ua->request($request, \&callback, 4096);
 
- sub callback { my($data, $response, $protocol) = @_; .... }
+ sub callback { my($data, $response, $protocol) = @_; ... }
 
 =head1 DESCRIPTION
 
@@ -44,7 +58,7 @@ request() methods, which dispatches it using the relevant protocol,
 and returns a C<HTTP::Response> object.
 
 There are convenience methods for sending the most common request
-types; get(), head() and post().
+types: get(), head() and post().
 
 The basic approach of the library is to use HTTP style communication
 for all protocol schemes, i.e. you even receive an C<HTTP::Response>
@@ -74,7 +88,7 @@ The subroutine variant requires a reference to callback routine as the
 second argument to the request method and it can also take an optional
 chuck size as the third argument.  This variant can be used to
 construct "pipe-lined" processing, where processing of received
-chuncks can begin before the complete data has arrived.  The callback
+chunks can begin before the complete data has arrived.  The callback
 function is called with 3 arguments: the data received this time, a
 reference to the response object and a reference to the protocol
 object.  The response object returned from the request method will
@@ -103,7 +117,7 @@ use vars qw(@ISA $VERSION);
 
 require LWP::MemberMixin;
 @ISA = qw(LWP::MemberMixin);
-$VERSION = sprintf("%d.%03d", q$Revision: 2.0 $ =~ /(\d+)\.(\d+)/);
+$VERSION = sprintf("%d.%03d", q$Revision: 2.4 $ =~ /(\d+)\.(\d+)/);
 
 use HTTP::Request ();
 use HTTP::Response ();
@@ -281,14 +295,14 @@ sub send_request
 
     my($method, $url) = ($request->method, $request->uri);
 
-    local($SIG{__DIE__});  # protect agains user defined die handlers
+    local($SIG{__DIE__});  # protect against user defined die handlers
 
     # Check that we have a METHOD and a URL first
-    return HTTP::Response->new(&HTTP::Status::RC_BAD_REQUEST, "Method missing")
+    return _new_response($request, &HTTP::Status::RC_BAD_REQUEST, "Method missing")
 	unless $method;
-    return HTTP::Response->new(&HTTP::Status::RC_BAD_REQUEST, "URL missing")
+    return _new_response($request, &HTTP::Status::RC_BAD_REQUEST, "URL missing")
 	unless $url;
-    return HTTP::Response->new(&HTTP::Status::RC_BAD_REQUEST, "URL must be absolute")
+    return _new_response($request, &HTTP::Status::RC_BAD_REQUEST, "URL must be absolute")
 	unless $url->scheme;
 
     LWP::Debug::trace("$method $url");
@@ -309,7 +323,7 @@ sub send_request
       #  into class LWP::Protocol::nogo.
       my $x;
       if($x       = $self->protocols_allowed) {
-        if(grep $_ eq $scheme, @$x) {
+        if(grep lc($_) eq $scheme, @$x) {
           LWP::Debug::trace("$scheme URLs are among $self\'s allowed protocols (@$x)");
         } else {
           LWP::Debug::trace("$scheme URLs aren't among $self\'s allowed protocols (@$x)");
@@ -317,7 +331,7 @@ sub send_request
           $protocol = LWP::Protocol::nogo->new;
         }
       } elsif ($x = $self->protocols_forbidden) {
-        if(grep $_ eq $scheme, @$x) {
+        if(grep lc($_) eq $scheme, @$x) {
           LWP::Debug::trace("$scheme URLs are among $self\'s forbidden protocols (@$x)");
           require LWP::Protocol::nogo;
           $protocol = LWP::Protocol::nogo->new;
@@ -332,8 +346,7 @@ sub send_request
       $protocol = eval { LWP::Protocol::create($scheme, $self) };
       if ($@) {
 	$@ =~ s/ at .* line \d+.*//s;  # remove file/line number
-
-	return HTTP::Response->new(&HTTP::Status::RC_NOT_IMPLEMENTED, $@);
+	return _new_response($request, &HTTP::Status::RC_NOT_IMPLEMENTED, $@);
       }
     }
 
@@ -546,6 +559,26 @@ This is a shortcut for C<$ua-E<gt>request(HTTP::Request::Common::GET(
 $url, Header =E<gt> Value,... ))>.  See
 L<HTTP::Request::Common|HTTP::Request::Common>.
 
+To specify a file that you want the content of the request to be
+saved to (instead of being saved in C<< $response->content >>),
+specify a header C<< ':content_file' => I<PATHSPEC> >>.
+This corresponds to C<< $ua->request( HTTP::Request::Commont::GET(...),
+'/path/to/that/file' ) >> (or whatever relative or absolute file
+pathspec you specify).
+
+To instead specify that the content, as it is received, should be sent
+to a routine, specify a header C<< ':content_cb' => I<SUBREF> >>. To
+suggest a size for the chunks of data sent to that callback, you can
+optionally specify a header C<< ':read_size_hint' => I<BYTECOUNT> >>.
+This corresponds to C<< $ua->request(
+HTTP::Request::Commont::GET(...), \&callback, 4096 ) >> (or whatever
+subref and optional bytecount you specify).
+
+These three optional headers that start with ":" are never
+made into real request headers, but are extracted and used as options
+to the eventual call to C<< $ua->request(REQ, options...) >> and
+thereby to C<< $ua->send_request(REQ, options...) >>.
+
 =item $ua->post($url, \%formref, Header => Value,...);
 
 This is a shortcut for C<$ua-E<gt>request( HTTP::Request::Common::POST(
@@ -556,38 +589,104 @@ reference is optional, and can be either a hashref (C<\%formdata> or C<{
 C<['key1' => 'val2', 'key2' => 'val2', ...]>).  See
 L<HTTP::Request::Common|HTTP::Request::Common>.
 
+You can also use the C<':content_file' / ':content_cb' /
+':read_size_hint'> headers just as with C<< $ua->get >>.
+
 =item $ua->head($url, Header => Value,...);
 
 This is a shortcut for C<$ua-E<gt>request( HTTP::Request::Common::HEAD(
 $url, Header =E<gt> Value,... ))>.  See
 L<HTTP::Request::Common|HTTP::Request::Common>.
 
-=item $ua->put($url, Header => Value,...);
-
-This is a shortcut for C<$ua-E<gt>request( HTTP::Request::Common::PUT(
-$url, Header =E<gt> Value,... ))>.  See
-L<HTTP::Request::Common|HTTP::Request::Common>.
+You can also use the C<':content_file' / ':content_cb' /
+':read_size_hint'> headers just as with C<< $ua->get >>.
 
 =cut
 
+#=item $ua->put($url, Header => Value,...);
+#
+#This is a shortcut for C<$ua-E<gt>request( HTTP::Request::Common::PUT(
+#$url, Header =E<gt> Value,... ))>.  See
+#L<HTTP::Request::Common|HTTP::Request::Common>.
+#
+#=cut
+
 sub get {
-  require HTTP::Request::Common;
-  return shift->request( HTTP::Request::Common::GET( @_ ) );
+    require HTTP::Request::Common;
+    my($self, @parameters) = @_;
+    my @suff = $self->_process_colonic_headers(\@parameters,1);
+    return $self->request( HTTP::Request::Common::GET( @parameters ), @suff );
 }
 
 sub post {
-  require HTTP::Request::Common;
-  return shift->request( HTTP::Request::Common::POST( @_ ) );
+    require HTTP::Request::Common;
+    my($self, @parameters) = @_;
+    my @suff = $self->_process_colonic_headers(\@parameters,2);
+    return $self->request( HTTP::Request::Common::POST( @parameters ), @suff );
 }
 
 sub head {
-  require HTTP::Request::Common;
-  return shift->request( HTTP::Request::Common::HEAD( @_ ) );
+    require HTTP::Request::Common;
+    my($self, @parameters) = @_;
+    my @suff = $self->_process_colonic_headers(\@parameters,1);
+    return $self->request( HTTP::Request::Common::HEAD( @parameters ), @suff );
 }
 
-sub put {
-  require HTTP::Request::Common;
-  return shift->request( HTTP::Request::Common::PUT( @_ ) );
+#sub put {
+#  require HTTP::Request::Common;
+#  my($self, @parameters) = @_;
+#  my @suff = $self->_process_colonic_headers(\@parameters,1);
+#  return $self->request( HTTP::Request::Common::PUT( @parameters ), @suff );
+#}
+
+
+sub _process_colonic_headers {
+    # Process :content_cb / :content_file / :read_size_hint headers.
+    my($self, $args, $start_index) = @_;
+
+    my($arg, $size);
+    for(my $i = $start_index; $i < @$args; $i += 2) {
+	next unless defined $args->[$i];
+
+	#printf "Considering %s => %s\n", $args->[$i], $args->[$i + 1];
+
+	if($args->[$i] eq ':content_cb') {
+	    # Some sanity-checking...
+	    $arg = $args->[$i + 1];
+	    Carp::croak("A :content_cb value can't be undef") unless defined $arg;
+	    Carp::croak("A :content_cb value must be a coderef")
+		unless ref $arg and UNIVERSAL::isa($arg, 'CODE');
+	    
+	}
+	elsif ($args->[$i] eq ':content_file') {
+	    $arg = $args->[$i + 1];
+
+	    # Some sanity-checking...
+	    Carp::croak("A :content_file value can't be undef")
+		unless defined $arg;
+	    Carp::croak("A :content_file value can't be a reference")
+		if ref $arg;
+	    Carp::croak("A :content_file value can't be \"\"")
+		unless length $arg;
+
+	}
+	elsif ($args->[$i] eq ':read_size_hint') {
+	    $size = $args->[$i + 1];
+	    # Bother checking it?
+
+	}
+	else {
+	    next;
+	}
+	splice @$args, $i, 2;
+	$i -= 2;
+    }
+
+    # And return a suitable suffix-list for request(REQ,...)
+
+    return             unless defined $arg;
+    return $arg, $size if     defined $size;
+    return $arg;
 }
 
 
@@ -598,16 +697,16 @@ sub put {
 
 =item $ua->protocols_allowed( \@protocols ); # to set
 
-This reads (or sets) this user-agent's list of procotols that
+This reads (or sets) this user agent's list of protocols that
 C<$ua-E<gt>request> and C<$ua-E<gt>simple_request> will exclusively
-allow.
+allow.  The protocol names are case insensitive.
 
 For example: C<$ua-E<gt>protocols_allowed( [ 'http', 'https'] );>
 means that this user agent will I<allow only> those protocols,
-and attempts to use this user-agent to access URLs with any other
+and attempts to use this user agent to access URLs with any other
 schemes (like "ftp://...") will result in a 500 error.
 
-To delete the list, call: 
+To delete the list, call:
 C<$ua-E<gt>protocols_allowed(undef)>
 
 By default, an object has neither a protocols_allowed list, nor
@@ -620,20 +719,21 @@ list causes any protocols_forbidden list to be ignored.
 
 =item $ua->protocols_forbidden( \@protocols ); # to set
 
-This reads (or sets) this user-agent's list of procotols that
+This reads (or sets) this user agent's list of protocols that
 C<$ua-E<gt>request> and C<$ua-E<gt>simple_request> will I<not> allow.
+The protocol names are case insensitive.
 
 For example: C<$ua-E<gt>protocols_forbidden( [ 'file', 'mailto'] );>
-means that this user-agent will I<not> allow those protocols, and
-attempts to use this user-agent to access URLs with those schemes
+means that this user agent will I<not> allow those protocols, and
+attempts to use this user agent to access URLs with those schemes
 will result in a 500 error.
 
-To delete the list, call: 
+To delete the list, call:
 C<$ua-E<gt>protocols_forbidden(undef)>
 
 =item $ua->is_protocol_supported($scheme)
 
-You can use this method to test whether this user-agent object supports the
+You can use this method to test whether this user agent object supports the
 specified C<scheme>.  (The C<scheme> might be a string (like 'http' or
 'ftp') or it might be an URI object reference.)
 
@@ -658,12 +758,12 @@ sub is_protocol_supported
 
     my $x;
     if(ref($self) and $x       = $self->protocols_allowed) {
-      return 0 unless grep $_ eq $scheme, @$x;
+      return 0 unless grep lc($_) eq $scheme, @$x;
     } elsif (ref($self) and $x = $self->protocols_forbidden) {
-      return 0 if grep $_ eq $scheme, @$x;
+      return 0 if grep lc($_) eq $scheme, @$x;
     }
 
-    local($SIG{__DIE__});  # protect agains user defined die handlers
+    local($SIG{__DIE__});  # protect against user defined die handlers
     $x = LWP::Protocol::implementor($scheme);
     return 1 if $x and $x ne 'LWP::Protocol::nogo';
     return 0;
@@ -755,7 +855,7 @@ This is called by request() to retrieve credentials for a Realm
 protected by Basic Authentication or Digest Authentication.
 
 Should return username and password in a list.  Return undef to abort
-the authentication resolution atempts.
+the authentication resolution attempts.
 
 This implementation simply checks a set of pre-stored member
 variables. Subclasses can override this method to e.g. ask the user
@@ -800,7 +900,7 @@ Examples are:
 =item $ua->_agent
 
 Returns the default agent identifier.  This is a string of the form
-"libwww-perl/#.##", where "#.##" is substitued with the version numer
+"libwww-perl/#.##", where "#.##" is substituted with the version number
 of this library.
 
 =cut
@@ -849,7 +949,7 @@ The default is to have no cookie_jar, i.e. never automatically add
 
 Shortcut: If a reference to a plain hash is passed in as the
 $cookie_jar_object, then it is replaced with an instance of
-C<HTTP::Cookies> that is initalized based on the hash.  This form also
+C<HTTP::Cookies> that is initialized based on the hash.  This form also
 automatically loads the C<HTTP::Cookies> module.  It means that:
 
   $ua->cookie_jar({ file => "$ENV{HOME}/.cookies.txt" });
@@ -865,7 +965,7 @@ Get/set the I<LWP::ConnCache> object to use.
 
 =item $ua->parse_head([$boolean])
 
-Get/set a value indicating wether we should initialize response
+Get/set a value indicating whether we should initialize response
 headers from the E<lt>head> section of HTML documents. The default is
 TRUE.  Do not turn this off, unless you know what you are doing.
 
@@ -1046,10 +1146,10 @@ specify proxies like this (sh-syntax):
   no_proxy="localhost,my.domain"
   export gopher_proxy wais_proxy no_proxy
 
-Csh or tcsh users should use the C<setenv> command to define these
+csh or tcsh users should use the C<setenv> command to define these
 environment variables.
 
-On systems with case-insensitive environment variables there exists a
+On systems with case insensitive environment variables there exists a
 name clash between the CGI environment variables and the C<HTTP_PROXY>
 environment variable normally picked up by env_proxy().  Because of
 this C<HTTP_PROXY> is not honored for CGI scripts.  The
@@ -1125,6 +1225,14 @@ sub _need_proxy
     undef;
 }
 
+sub _new_response {
+    my($request, $code, $message) = @_;
+    my $response = HTTP::Response->new($code, $message);
+    $response->request($request);
+    $response->header("Client-Date" => HTTP::Date::time2str(time));
+    return $response;
+}
+
 1;
 
 =back
@@ -1136,7 +1244,7 @@ F<lwp-mirror> for examples of usage.
 
 =head1 COPYRIGHT
 
-Copyright 1995-2001 Gisle Aas.
+Copyright 1995-2002 Gisle Aas.
 
 This library is free software; you can redistribute it and/or
 modify it under the same terms as Perl itself.
